from typing import TypeVar, Annotated

from flask_typed.parameter import ParameterLocation

_T = TypeVar("T")

Query = Annotated[_T, ParameterLocation.QUERY]
Path = Annotated[_T, ParameterLocation.PATH]
Header = Annotated[_T, ParameterLocation.HEADER]



from collections import defaultdict
from inspect import isclass
from types import UnionType
from typing import get_origin, get_args, NamedTuple, Type

import openapi_schema_pydantic as openapi
from openapi_schema_pydantic.util import PydanticSchema
from pydantic import BaseModel

from flask_typed.errors import HttpError
from flask_typed.response import BaseResponse


class ResponseInfo(NamedTuple):
    schema: openapi.Schema
    description: str | None


class ResponsesDocsBuilder:

    def __init__(self, return_type, docstring, docs):
        self.responses = defaultdict(lambda: defaultdict(list))
        self.return_type = return_type
        self.docs = docs
        self.docstring = docstring

    def build(self) -> dict[str, openapi.Response]:
        origin_type = get_origin(self.return_type)
        if origin_type is UnionType:
            types = get_args(self.return_type)
        else:
            types = [self.return_type]

        for response_type in types:
            self._add_response(response_type)

        if self.docs:
            for error_model in self.docs.errors:
                self._add_response(error_model)

        return self._merge_responses()

    def _add_response(self, response_type):
        if isclass(response_type):
            description = self._get_return_description(response_type)
            if issubclass(response_type, BaseModel):
                status_code = getattr(response_type.Config, "status_code", 200)
                self.responses[status_code]["application/json"].append(
                    ResponseInfo(
                        schema=PydanticSchema(schema_class=response_type),
                        description=description
                    )
                )
            elif issubclass(response_type, BaseResponse):
                status_code = response_type.status_code
                mime_type = response_type.mime_type
                schema = response_type.schema()
                if not isinstance(schema, openapi.Schema):
                    raise TypeError(
                        f"Response type {response_type} extending BaseResponse"
                        f" does not implement schema() method properly: {schema}"
                    )

                self.responses[status_code][mime_type].append(
                    ResponseInfo(
                        schema=schema,
                        description=description
                    )
                )
            elif issubclass(response_type, HttpError):
                self.responses[response_type.status_code]["application/json"].append(
                    ResponseInfo(
                        schema=response_type.schema(),
                        description=description
                    )
                )
        else:
            description = self._get_return_description(response_type.__class__)
            if isinstance(response_type, HttpError):
                self.responses[response_type.status_code]["application/json"].append(
                    ResponseInfo(
                        schema=response_type.schema(),
                        description=description
                    )
                )

    def _merge_responses(self) -> dict[str, openapi.Response]:
        response_docs = {}
        for status_code, responses_with_status in self.responses.items():
            contents = {}
            description = None
            for mime_type, response_info_list in responses_with_status.items():
                if len(response_info_list) == 1:
                    schema = response_info_list[0].schema
                else:
                    schema = openapi.Schema(
                        oneOf=[response_info.schema for response_info in response_info_list]
                    )
                for response_info in response_info_list:
                    if description:
                        break
                    description = response_info.description

                contents[mime_type] = openapi.MediaType(schema=schema)

            # If no description is found for specific return type, use generic return docstring for successes
            if not description and 200 <= status_code < 300:
                description = self._get_return_description(None)

            response_docs[str(status_code)] = openapi.Response(
                description=description if description else "",
                content=contents
            )

        return response_docs

    def _get_return_description(self, return_type: Type | None) -> str | None:
        if self.docstring:
            key = return_type.__name__ if return_type is not None else None
            if returns := self.docstring.returns.get(key):
                return returns.description
            if raises := self.docstring.raises.get(key):
                return raises.description
        return None



from abc import ABC
from typing import TypeVar, Generic

import openapi_schema_pydantic as openapi
from flask import Request
from werkzeug.datastructures import MultiDict, Headers


T = TypeVar("T")


class RequestParser(ABC):

    @classmethod
    def parse_request(cls, request: Request) -> 'Self':
        raise NotImplementedError


class QueryParser(RequestParser, ABC):

    @classmethod
    def parse(cls, args: MultiDict[str, str]) -> 'Self':
        raise NotImplementedError

    @classmethod
    def schema(cls) -> list[openapi.Parameter]:
        raise NotImplementedError

    @classmethod
    def parse_request(cls, request: Request) -> 'Self':
        return cls.parse(request.args)


class HeaderParser(RequestParser, ABC):

    @classmethod
    def parse(cls, header: Headers) -> 'Self':
        raise NotImplementedError

    @classmethod
    def schema(cls) -> list[openapi.Parameter]:
        raise NotImplementedError

    @classmethod
    def parse_request(cls, request: Request) -> 'Self':
        return cls.parse(request.headers)


class BodyParser(RequestParser, ABC):

    @classmethod
    def parse(cls, data: bytes) -> 'Self':
        raise NotImplementedError

    @classmethod
    def schema(cls) -> openapi.RequestBody:
        raise NotImplementedError

    @classmethod
    def parse_request(cls, request: Request) -> 'Self':
        return cls.parse(request.data)



import inspect
from inspect import isclass
from typing import Any, get_origin, Annotated, get_args, Type

import openapi_schema_pydantic as openapi
from flask import request, current_app
from pydantic import BaseModel

from flask_typed.docs.responses import ResponsesDocsBuilder
from flask_typed.docs.utils import Docstring
from .errors import HttpError
from .parameter import ParameterLocation, Parameter, ParameterValidationError, ValidationError
from .parsers import RequestParser
from .response import BaseResponse


class HttpHandler:

    def __init__(self, path, resource_cls, handler):
        self.resource_cls = resource_cls
        self.path = path
        self.handler = handler
        docstring = getattr(handler, "__doc__", None)
        self.docstring = Docstring(docstring) if docstring else None
        self.docs_metadata = getattr(handler, "docs_metadata", None)
        self.responses = None
        self.parameters: list[Parameter] = []
        self.request_parsers: dict[str, Type[RequestParser]] = {}

        self._process_annotations()

    def _process_annotations(self):
        handler_signature = inspect.signature(self.handler)
        for parameter in handler_signature.parameters.values():
            if parameter.name == "self":
                continue

            source_name = parameter.name
            param_type = parameter.annotation
            if param_type is None:
                raise TypeError(f"No type annotation is provided for parameter: {parameter.name}")

            if isclass(param_type) and issubclass(param_type, RequestParser):
                self.request_parsers[source_name] = param_type
                continue

            if get_origin(param_type) == Annotated:
                metadata = get_args(param_type)
                param_type = metadata[0]
                location = metadata[1]

                if location == ParameterLocation.HEADER:
                    source_name = "-".join(
                        s.capitalize() for s in parameter.name.split("_")
                    )
                if len(metadata) > 2:
                    source_name = metadata[2]

            elif isclass(param_type) and issubclass(param_type, BaseModel):
                location = ParameterLocation.BODY
            elif parameter.name in self.path.path_parameters:
                location = ParameterLocation.PATH
            else:
                location = ParameterLocation.QUERY

            match parameter.default:
                case parameter.empty:
                    default_value = ...
                case other:
                    default_value = other

            parameter = Parameter(
                name=parameter.name,
                source=source_name,
                location=location,
                param_type=param_type,
                description=self._get_parameter_description(parameter.name),
                default_value=default_value
            )

            self.parameters.append(parameter)

        self.responses = ResponsesDocsBuilder(
            return_type=handler_signature.return_annotation,
            docstring=self.docstring,
            docs=self.docs_metadata
        ).build()

    def generate_operation(self) -> openapi.Operation:
        doc_parameters = []
        request_body = None
        for param in self.parameters:
            match param.location:
                case ParameterLocation.QUERY | ParameterLocation.PATH | ParameterLocation.HEADER:
                    doc_parameters.extend(param.to_openapi_parameters())
                case ParameterLocation.BODY:
                    request_body = param.to_openapi_request_body()

        return openapi.Operation(
            parameters=doc_parameters,
            responses=self.responses,
            requestBody=request_body,
            summary=self.docstring.short_description if self.docstring else "",
            description=self.docstring.long_description if self.docstring else "",
        )

    def get_handler(self):
        parameters = self.parameters
        parsers = self.request_parsers
        handler = self.handler
        resource_cls = self.resource_cls

        def perform_validation(kwargs) -> dict[str, Any]:
            validated_args = {}
            validation_errors = []

            for param in parameters:
                try:
                    validated_args[param.name] = param.validate(request, kwargs)
                except ParameterValidationError as e:
                    validation_errors.append(e)

            for name, parser in parsers.items():
                try:
                    validated_args[name] = parser.parse_request(request)
                except ParameterValidationError as e:
                    validation_errors.append(e)

            if validation_errors:
                raise ValidationError(errors=[
                    err.to_model() for err in validation_errors
                ])

            return validated_args

        def validated(*_args, **kwargs):
            try:
                validated_args = perform_validation(kwargs)
            except HttpError as e:
                return e.flask_response()

            try:
                response_value = handler(resource_cls, **validated_args)
            except HttpError as e:
                return e.flask_response()

            if isinstance(response_value, BaseResponse):
                return response_value.flask_response()
            if isinstance(response_value, BaseModel):
                return current_app.response_class(
                    response=response_value.json(),
                    status=getattr(response_value.Config, "status_code", 200),
                    mimetype='application/json',
                )
            else:
                return response_value

        return validated

    def _get_parameter_description(self, param_name) -> str:
        return "" if self.docstring is None else self.docstring.get_parameter_description(param_name)



import json
from typing import Type

from flask import Flask, render_template_string
from openapi_schema_pydantic import OpenAPI, Info
from openapi_schema_pydantic.util import construct_open_api_with_schema_class

from flask_typed.docs.utils import redoc_template
from .typed_resource import BoundResource, TypedResource


def join_path(path1: str, path2: str) -> str:
    return f"{path1.rstrip('/')}/{path2.lstrip('/')}"


class TypedBlueprint:

    def __init__(self):
        self.resources = {}

    def add_resource(self, resource: Type[TypedResource], path: str):
        self.resources[path] = resource


class TypedAPI:

    def __init__(
            self,
            app: Flask | None = None,
            version: str = "v0.0.1",
            description: str = "",
            openapi_path: str = "/openapi",
            docs_path: str = "/docs"
     ):
        self.app = app
        self.docs = OpenAPI(
            info=Info(
                title=description,
                version=version
            ),
            paths={}
        )
        self.resources: dict[str, BoundResource] = {}
        self.openapi_path = openapi_path
        self.docs_path = docs_path

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        if app is None:
            raise ValueError("No valid Flask instance is provided")
        self.app = app
        for url, resource in self.resources.items():
            self._register_resource(resource)

        def get_openapi_schema():
            return self.get_openapi_schema()

        def redoc():
            return render_template_string(
                redoc_template,
                api_doc_url=self.openapi_path
            )

        app.add_url_rule(self.docs_path, view_func=redoc)
        app.add_url_rule(self.openapi_path, view_func=get_openapi_schema)

    def add_resource(self, resource: Type[TypedResource], path: str):
        if path in self.resources:
            raise Exception(f"URL is already registered: {path}")
        bound_resource = resource.bind(path)
        self.resources[path] = bound_resource
        self.docs.paths[bound_resource.path.openapi_path] = bound_resource.generate_path_item()

        if self.app is not None:
            self._register_resource(bound_resource)

    def register_blueprint(self, blueprint: TypedBlueprint, url_prefix: str = ""):
        for path, resource in blueprint.resources.items():
            full_path = join_path(url_prefix, path)
            self.add_resource(resource, full_path)

    def _register_resource(self, bound_resource: BoundResource):
        for method, handler in bound_resource.methods.items():
            self.app.add_url_rule(
                bound_resource.path.path,
                bound_resource.resource_cls.__name__.lower() + method,
                handler.get_handler(),
                methods=[method]
            )

    def get_openapi_schema(self):
        open_api = construct_open_api_with_schema_class(self.docs)
        return json.loads(open_api.json(by_alias=True, exclude_none=True))



import builtins
from typing import Type

import docstring_parser

from flask_typed.errors import HttpError

_builtin_openapi_map = {
    builtins.bool: "boolean",
    builtins.str: "string",
    builtins.int: "integer",
    builtins.float: "number"
}


def get_builtin_type(ty: Type):
    return _builtin_openapi_map.get(ty)


class DocsMetadata:

    def __init__(
            self,
            errors: list[HttpError | Type[HttpError]] | None = None
    ):
        self.errors = errors if errors is not None else []


def docs(
        errors: list[HttpError | Type[HttpError]] | None = None
):
    def docs_decorator(func):
        func.docs_metadata = DocsMetadata(
            errors=errors
        )
        return func

    return docs_decorator


class Docstring:

    def __init__(self, docstring: str):
        docstring = docstring_parser.parse(docstring)
        self.short_description = docstring.short_description
        self.long_description = docstring.long_description
        self.params = {param.arg_name: param for param in docstring.params}
        self.returns = {returns.type_name: returns for returns in docstring.many_returns}
        self.raises = {exception.type_name: exception for exception in docstring.raises}

    def get_parameter_description(self, name: str) -> str:
        if parameter := self.params.get(name):
            return parameter.description
        return ""


redoc_template = """
<!DOCTYPE html>
<html>
  <head>
    <title>Redoc</title>
    <!-- needed for adaptive design -->
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">

    <!--
    Redoc doesn't change outer page styles
    -->
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <redoc spec-url='{{ api_doc_url }}'></redoc>
    <script src="https://cdn.redoc.ly/redoc/latest/bundles/redoc.standalone.js"> </script>
  </body>
</html>
"""


from abc import ABC

import openapi_schema_pydantic as openapi
from flask import current_app
from pydantic import BaseModel

from .response import BaseResponse


class BaseHttpError(Exception, BaseResponse, ABC):
    pass


class HttpError(BaseHttpError):
    mime_type = "application/json"

    class ResponseModel(BaseModel):
        message: str | None = None

    def __init_subclass__(cls, **kwargs):
        if not issubclass(cls.ResponseModel, BaseModel):
            raise TypeError(f"ResponseModel should inherit pydantic BaseModel: {cls.__name__}")

    def __init__(self, status_code: int | None = None, **kwargs):
        cls = self.__class__
        self.status_code = cls.status_code if status_code is None else status_code
        self.response = cls.ResponseModel(**kwargs)

    def flask_response(self):
        return current_app.response_class(
            response=self.json(),
            status=self.status_code,
            mimetype=self.mime_type,
        )

    def json(self) -> str:
        return self.response.json()

    @classmethod
    def schema(cls) -> openapi.Schema:
        return openapi.Schema.parse_obj(cls.ResponseModel.schema())


class MessageHttpError(HttpError):
    message: str = "Error"

    def __init_subclass__(cls, **kwargs):
        class ResponseModel(BaseModel):
            message: str | None = cls.message

        cls.ResponseModel = ResponseModel


class BadRequestError(MessageHttpError):
    status_code = 400
    message = "Bad request"


class NotFoundError(HttpError):
    status_code = 404
    message = "Not found"


class InternalServerError(HttpError):
    status_code = 500
    message = "Internal server error"



import re

import openapi_schema_pydantic as openapi
from flask.views import http_method_funcs

from .handler import HttpHandler

_PATH_REGEX = re.compile("<(?:(?P<converter>[A-Za-z_]\\w*):)?(?P<name>[A-Za-z_]\\w*)>")


def parse_path_for_parameters(path) -> dict[str, str]:
    m = _PATH_REGEX.findall(path)
    return {name: converter for converter, name in m}


def convert_to_openapi_format(path) -> str:
    return _PATH_REGEX.sub("{\\g<name>}", path)


class Path:

    def __init__(self, path: str):
        self.path = path
        self.path_parameters = parse_path_for_parameters(path)
        self.openapi_path = convert_to_openapi_format(path)


class BoundResource:

    def __init__(self, resource_cls, path: Path, methods: dict[str, HttpHandler]):
        self.resource_cls = resource_cls
        self.path = path
        self.methods = methods

    def generate_path_item(self) -> openapi.PathItem:
        docs = openapi.PathItem()
        for method, handler in self.methods.items():
            operation = handler.generate_operation()
            setattr(docs, method.lower(), operation)
        return docs


class TypedResource:

    @classmethod
    def bind(cls, path: str) -> BoundResource:
        path = Path(path)
        methods = {}
        for method in http_method_funcs:
            if handler_method := getattr(cls, method, None):
                handler = HttpHandler(path, cls, handler_method)
                methods[method.upper()] = handler

        return BoundResource(
            resource_cls=cls,
            path=path,
            methods=methods
        )



from typing import Generator, Any

import openapi_schema_pydantic as openapi
from flask import current_app, stream_with_context
from openapi_schema_pydantic.util import PydanticSchema
from pydantic import BaseModel


_pydantic_export_config_fields = [
    "include",
    "exclude",
    "by_alias",
    "exclude_unset",
    "exclude_defaults",
    "exclude_none"
]


class BaseResponseConfig:
    mime_type: str = "text/plain"
    status_code: int = 200


class BaseResponse:

    class ResponseConfig:
        pass

    def __init_subclass__(cls, **kwargs):
        response_config = cls.ResponseConfig
        cls._mime_type = getattr(response_config, "mime_type", BaseResponseConfig.mime_type)
        cls._status_code = getattr(response_config, "status_code", BaseResponseConfig.status_code)

    def flask_response(self):
        raise NotImplementedError

    @classmethod
    def schema(cls) -> openapi.Schema:
        return openapi.Schema(type="string")


class ModelResponse(BaseModel, BaseResponse):

    def __init_subclass__(cls, **kwargs):
        export_config = {}

        for field in _pydantic_export_config_fields:
            config = getattr(cls.ResponseConfig, field, ...)
            if config is not ...:
                export_config[field] = config

        cls._export_config = export_config

        super().__init_subclass__(**kwargs)

    def flask_response(self):
        return current_app.response_class(
            response=self.json(**self._export_config) if self._export_config else self.json(),
            mimetype=self._mime_type,
            status=self._status_code
        )

    @classmethod
    def schema(cls) -> openapi.Schema:
        return PydanticSchema(schema_class=cls)


class Response(BaseResponse):

    def __init__(self, body: str):
        self._body = body

    def flask_response(self):
        return current_app.response_class(
            response=self._body,
            mimetype=self.mime_type,
            status=self.status_code
        )


class StreamingResponse(BaseResponse):
    status_code = 200
    mime_type = "text/plain"

    def __init__(self, generator: Generator, use_context=True):
        self._generator = generator
        self._use_context = use_context

    def flask_response(self):
        return current_app.response_class(
            response=stream_with_context(self._generator) if self._use_context else self._generator,
            mimetype=self.mime_type,
            status=self.status_code
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_typed', 'flask_typed.docs']

package_data = \
{'': ['*']}

install_requires = \
['docstring-parser>=0.15,<0.16',
 'flask>=2.2.3,<3.0.0',
 'openapi-schema-pydantic>=1.2.4,<2.0.0',
 'pydantic>=1.10.5,<2.0.0']

setup_kwargs = {
    'name': 'flask-typed',
    'version': '0.1.0',
    'description': '',
    'long_description': '# flask-typed\n\nA Flask extension for developing HTTP APIs using type annotations. Type annotations are used for the validation of requests and generating API documentation.\n\n## Example\n\n```python\nfrom flask import Flask\nfrom pydantic import BaseModel\n\nfrom flask_typed import TypedResource, TypedAPI\n\n\nclass HelloResponse(BaseModel):\n\n    message: str\n    sender: str\n    receiver: str\n\n\nclass HelloResource(TypedResource):\n\n    def get(self, sender: str, receiver: str) -> HelloResponse:\n        """\n        Greets someone\n\n        :param sender: Greeter\n        :param receiver: The one being greeted\n        :return: Greetings\n        """\n        return HelloResponse(\n            message=f"Hello to {receiver} from {sender}!",\n            sender=sender,\n            receiver=receiver\n        )\n\n\napp = Flask(__name__)\napi = TypedAPI(app, version="1.0", description="Greetings API")\n\napi.add_resource(HelloResource, "/hello/<sender>")\n\nif __name__ == "__main__":\n    app.run()\n```',
    'author': 'Mustafa Efendioğlu',
    'author_email': 'mfnd@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import json
from enum import IntEnum
from inspect import isclass
from types import UnionType, NoneType
from typing import Type, Any, get_origin, get_args, Sequence

import openapi_schema_pydantic as openapi
import pydantic
from openapi_schema_pydantic.util import PydanticSchema
from pydantic import BaseModel

from flask_typed.docs.utils import get_builtin_type
from .errors import HttpError
from .parsers import QueryParser, HeaderParser


class ParameterLocation(IntEnum):
    PATH = 1
    QUERY = 2
    BODY = 3
    HEADER = 4


class Parameter:

    def __init__(
            self,
            name: str,
            source: str,
            location: ParameterLocation,
            param_type: Type,
            description: str,
            default_value: Any,
    ):
        self.name = name
        self.source = source
        self.location = location
        self.description = description
        self.is_optional = False
        self.type = param_type
        self.default_value = default_value

        self._init_types(param_type)
        self._init_data_getter()
        self._init_validator(self.type)

    def _init_types(self, param_type: Type):
        actual_type = None
        origin_type = get_origin(param_type)
        if origin_type is UnionType:
            for alternate_type in get_args(param_type):
                if alternate_type is NoneType:
                    continue
                if actual_type is not None:
                    raise Exception("Multiple argument types are provided")
                actual_type = alternate_type
        else:
            actual_type = param_type

        self.type = actual_type

        if self.default_value is Ellipsis:
            self.is_optional = False
        else:
            self.is_optional = True

    def _init_data_getter(self):
        match self.location:
            case ParameterLocation.QUERY:
                def get_query_param(request, _path_params):
                    return request.args.get(self.source)
                self.get_data = get_query_param
            case ParameterLocation.HEADER:
                def get_header_param(request, _path_params):
                    return request.headers.get(self.source)
                self.get_data = get_header_param
            case ParameterLocation.PATH:
                def get_path_param(_request, path_params):
                    return path_params.get(self.source)
                self.get_data = get_path_param
            case ParameterLocation.BODY:
                def get_body_param(request, _path_params):
                    return request.data
                self.get_data = get_body_param
            case _:
                raise ValueError(f"Invalid parameter location: {self.location}")

    def _init_validator(self, param_type):
        if issubclass(param_type, BaseModel):
            def model_validator(value):
                return param_type.parse_obj(
                    json.loads(value)
                )
            self.validator = model_validator
        else:
            self.validator = param_type

    def validate(self, request, path_params):
        value = self.get_data(request, path_params)
        if value is None:
            if self.is_optional is True:
                return self.default_value
            else:
                raise ParameterValidationError(self, errors=["Parameter is not optional"])
        try:
            return self.validator(value)
        except pydantic.ValidationError as e:
            raise ParameterValidationError(self, errors=[e])
        except Exception as e:
            raise ParameterValidationError(self, errors=[str(e)])

    def to_openapi_parameters(self) -> list[openapi.Parameter]:
        location = self.location.name.lower()
        parameters = []
        if isclass(self.type) and issubclass(self.type, BaseModel):
            query_schema = self.type.schema_json()
            for name, prop in query_schema["properties"].items():
                parameters.append(
                    openapi.Parameter(
                        name=name,
                        param_in=location,
                        param_schema=openapi.Schema.parse_obj(prop)
                    )
                )
        elif isclass(self.type) and issubclass(self.type, (QueryParser, HeaderParser)):
            parameters.extend(self.type.schema())
        else:
            openapi_type = get_builtin_type(self.type)
            if openapi_type is None:
                raise TypeError(f"Unsupported type for parameter '{self.name}': {self.type}")

            schema = openapi.Schema(type=openapi_type)
            if self.default_value is not Ellipsis:
                schema.default = self.default_value
            parameters.append(
                openapi.Parameter(
                    name=self.source,
                    description=self.description,
                    param_in=location,
                    param_schema=schema,
                    required=not self.is_optional
                )
            )
        return parameters

    def to_openapi_request_body(self) -> openapi.RequestBody:
        if isclass(self.type) and issubclass(self.type, BaseModel):
            schema = PydanticSchema(schema_class=self.type)
        else:
            openapi_type = get_builtin_type(self.type)
            if openapi_type is None:
                raise TypeError(f"Unsupported type for parameter '{self.name}': {self.type}")

            schema = openapi_type

        return openapi.RequestBody(
            content={
                "application/json": openapi.MediaType(
                    schema=schema
                )
            }
        )


def repr_pydantic_validation_error(err: pydantic.ValidationError) -> Sequence[str]:
    for error in err.errors():
        yield f"{error['msg']}: {'.'.join(error['loc'])}"


class ParameterValidationErrorModel(BaseModel):
    parameter: str
    location: str
    details: list[str]


class ParameterValidationError(Exception):

    def __init__(self, parameter: Parameter, errors: list[str | pydantic.ValidationError]):
        self.parameter = parameter
        self.errors = errors

    def to_model(self):
        details = []
        for error in self.errors:
            if isinstance(error, pydantic.ValidationError):
                details.extend(repr_pydantic_validation_error(error))
            else:
                details.append(error)

        return ParameterValidationErrorModel(
            parameter=self.parameter.source,
            location=self.parameter.location.name.lower(),
            details=details
        )


class ValidationError(HttpError):
    status_code = 422

    class ResponseModel(BaseModel):
        errors: list[ParameterValidationErrorModel]
